//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  Copyright(c)  2006  Analogic Corporation	(rjohnson@analogic.com)
//
//  This program may be distributed under the GNU Public License
//  version 2, as published by the Free Software Foundation, Inc.,
//  59 Temple Place, Suite 330 Boston, MA, 02111.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  Interactive 'emergency shell' that replaces /sbin/init for such
//  use. It provides for terminal control of the process and protects
//  against getting accidentally killed.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/ioctl.h>

#include "tty.h"

static void hangup(int sig);
static void reaper(int sig);

//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  Create and initialize a table for the signals to be initialized.
//  Add any signals I forgot to this table.
//
typedef struct {
    int signal;            // Signal value
    sig_t handler;         // Signal handler
    int flags;             // Signal flags
    } SIGS;

const SIGS sigs[] = {
    { SIGHUP,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGINT,    hangup,     SA_INTERRUPT|SA_RESTART },
    { SIGQUIT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGILL,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTRAP,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGABRT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGIOT,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGBUS,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGFPE,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGSEGV,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGUSR2,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGPIPE,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGALRM,   SIG_IGN,    SA_INTERRUPT|SA_RESTART },
    { SIGTERM,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGCHLD,   reaper,     SA_INTERRUPT|SA_RESTART },
    { SIGSTKFLT, SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGCONT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTSTP,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTTIN,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTTOU,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGURG,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGWINCH,  SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGIO,     SIG_DFL,    SA_INTERRUPT|SA_RESTART }};

#define ArraySize(a) (sizeof(a) / sizeof(a[0]))

static const char term[]="/dev/tty1";	// Alt-F1
static const char st[] = TERMIOS;

static void reaper(int sig)
{
    while(wait3(&sig, WNOHANG, NULL) > 0)
        ;
}
static void hangup(int sig)
{
    if(getpid() == 1)			// Keep init alive
        return;
    exit(EXIT_SUCCESS);
}

static void set_sig(int sig, sig_t funct, int flags)
{
    struct sigaction sa;
    if(sigaction(sig, NULL, &sa) < 0)
        return;
    sa.sa_flags = flags;
    sa.sa_handler = funct;
    (void)sigaction(sig, &sa, NULL);
}

int main(int count, char *argv[], char *env[])
{
    int fd, flags;
    size_t i;
    pid_t pid;
    for(;;)					// Run forever
    {
        switch((pid = fork()))
        {
        case 0:					// Child
            setsid();
            strcpy(argv[0], "/bin/bash");
            if((fd = open(term, O_RDWR|O_NDELAY)) < 0) {
                fd = 0x00;
            }
            else {
                flags = fcntl(fd, F_GETFL);
                flags &= ~O_NDELAY;
                fcntl(fd, F_SETFL, flags);
            }
            (void)tcsetattr(fd, TCSANOW, (struct termios *)st);
                                  		// Set interactive parameters
            fcntl(fd, F_SETOWN, getpid());	// Set terminal ownership
            ioctl(fd, TIOCSCTTY, 1);		// Set terminal control
            (void)dup2(fd, STDIN_FILENO);
            (void)dup2(fd, STDOUT_FILENO);
            (void)dup2(fd, STDERR_FILENO);
            for(i=0; i< ArraySize(sigs); i++)
                set_sig(sigs[i].signal, sigs[i].handler, sigs[i].flags);
            return execve(argv[0], argv, env);
       case -1:					// Failed
           exit(EXIT_FAILURE);
       default:					// Parent
           (void)waitpid(pid, &flags, 0);
        }
    }
    return 0;	// Not reached
}

